# Generated by Django 5.2 on 2025-04-25 20:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0014_softdeletecompositepk"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="ConcreteChild",
            fields=[
                (
                    "changedcondition_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.changedcondition",
                    ),
                ),
                ("child_field", models.CharField(max_length=16)),
            ],
            bases=("tests.changedcondition",),
        ),
        migrations.CreateModel(
            name="AbstractChild",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.CharField(max_length=16)),
                ("int_field", models.IntegerField(default=0)),
                ("dt_field", models.DateTimeField(auto_now=True)),
                ("nullable", models.CharField(default=None, max_length=16, null=True)),
                ("child_field", models.CharField(max_length=16)),
                (
                    "char_pk_fk_field",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="tests.charpk",
                    ),
                ),
                (
                    "fk_field",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "m2m_field",
                    models.ManyToManyField(related_name="+", to=settings.AUTH_USER_MODEL),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]
