# Generated by Django 3.2.15 on 2022-08-18 10:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0011_auto_20220817_2211"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="partitionmodel",
            options={},
        ),
    ]
