# Generated by Django 3.2.15 on 2022-08-11 11:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0008_searchmodel"),
    ]

    operations = [
        migrations.CreateModel(
            name="OrderSchema",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("int_field", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="ReceiptSchema",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("char_field", models.CharField(max_length=128)),
            ],
            options={
                "db_table": "table.with.dots",
            },
        ),
    ]
